from core.engine import get_lr
from core.utils import *


def train(
    model,
    train_loader,
    criterion,
    optimizer,
    scheduler,
    device,
    logger,
    tensorboard_logger,
    epoch,
    args,
    finetune=False,
):
    """
        Train one epoch for auxnet
        :param epoch: index of epoch
        """

    metric_logger = MetricLogger(logger=logger, delimiter="  ")
    metric_logger.add_meter("lr", SmoothedValue(window_size=1, fmt="{value}"))
    metric_logger.add_meter("img/s", SmoothedValue(window_size=10, fmt="{value}"))

    if finetune:
        model.eval()
        logger.info("Eval mode")
    else:
        model.train()
        logger.info("Train mode")

    header = "Epoch: [{}]".format(epoch)
    for image, target in metric_logger.log_every(
        train_loader, args.print_frequency, header
    ):
        start_time = time.time()
        image, target = image.to(device), target.to(device)

        # forward
        output = model(image)
        loss = criterion(output, target)

        # backward
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        acc1, acc5 = accuracy(output, target, topk=(1, 5))
        batch_size = image.shape[0]
        metric_logger.update(loss=loss.item(), lr=optimizer.param_groups[0]["lr"])
        metric_logger.meters["acc1"].update(acc1.item(), n=batch_size)
        metric_logger.meters["acc5"].update(acc5.item(), n=batch_size)
        metric_logger.meters["img/s"].update(
            batch_size * args.world_size / (time.time() - start_time)
        )

    # gather the stats from all processes
    metric_logger.synchronize_between_processes()

    scheduler.step()
    lr = get_lr(optimizer)
    logger.info("Change Learning rate: {}".format(lr))

    train_error = 100 - metric_logger.acc1.global_avg
    train_loss = metric_logger.loss.global_avg
    train5_error = 100 - metric_logger.acc5.global_avg
    if tensorboard_logger is not None:
        tensorboard_logger.add_scalar("train_top1_error", train_error, epoch)
        tensorboard_logger.add_scalar("train_top5_error", train5_error, epoch)
        tensorboard_logger.add_scalar("train_loss", train_loss, epoch)
        tensorboard_logger.add_scalar("lr", lr, epoch)
        for name, module in model.named_modules():
            if isinstance(module, (args.conv_type, args.fc_type)):
                if hasattr(module, "weight_clip_value"):
                    tensorboard_logger.add_scalar(
                        "{}_{}".format(name, "weight_clip_value"),
                        module.weight_clip_value,
                        epoch,
                    )
                if hasattr(module, "activation_clip_value"):
                    tensorboard_logger.add_scalar(
                        "{}_{}".format(name, "activation_clip_value"),
                        module.activation_clip_value,
                        epoch,
                    )

    logger.info(
        "|===>Training Error: {:.4f} Loss: {:.4f}, Top5 Error: {:.4f}".format(
            train_error, train_loss, train5_error
        )
    )
    return train_error, train_loss, train5_error

