import os

import torch
import torch.nn as nn
from torch.utils.tensorboard import SummaryWriter

import core.models as models
from core.checkpoint import CheckPoint
from core.config import copy_code, set_save_path
from core.dataloader import get_dataloader
from core.engine import train, val
from core.label_smooth import LabelSmoothCrossEntropyLoss
from core.logger import get_logger
from core.optim import get_optimizer, get_scheduler
from core.utils import *
from core.write_log import write_log, write_settings
from fpconfig import args

if __name__ == "__main__":
    # set gpu
    set_gpu(args)
    device = torch.device("cuda")

    # init distribution
    init_distributed_mode(args)
    set_save_path(args)
    if is_main_process():
        write_settings(args)
    if args.distributed:
        torch.distributed.barrier()

    # set logger
    logger = get_logger(args.save_path, "main")
    setup_logger_for_distributed(args.rank == 0, logger)

    # set tensorboard logger
    tensorboard_logger = SummaryWriter(args.save_path)
    setup_tensorboard_logger_for_distributed(args.rank == 0, tensorboard_logger)

    # backup code
    copy_code(
        logger, src=os.path.abspath("./"), dst=os.path.join(args.save_path, "code")
    )
    logger.info("|===>Result will be saved at {}".format(args.save_path))

    # get loader and model
    set_reproducible()
    train_loader, val_loader, train_sampler, val_sampler = get_dataloader(args, logger)
    model = models.__dict__[args.network](num_classes=args.n_classes)
    logger.info(model)

    # get checkpoint
    checkpoint = CheckPoint(args.save_path, logger)
    # load pretrained
    if args.pretrained is not None:
        check_point_params = torch.load(args.pretrained, map_location="cpu")
        model_state = check_point_params
        if args.network in ["mobilenetv1"]:
            model_state = check_point_params["model"]
        model = checkpoint.load_state(model, model_state)
        logger.info("|===>load restrain file: {}".format(args.pretrained))

    # load resume
    start_epoch = 0
    optimizer_state = None
    lr_scheduler_state = None
    if args.resume is not None:
        (
            model_state,
            optimizer_state,
            epoch,
            lr_scheduler_state,
        ) = checkpoint.load_checkpoint(args.resume)
        model = checkpoint.load_state(model, model_state)
        start_epoch = epoch
        optimizer_state = optimizer_state
        lr_scheduler_state = lr_scheduler_state
        logger.info("|===>load resume file: {}".format(args.resume))

    # move model to gpu
    model = model.to(device)
    if args.distributed:
        model = torch.nn.parallel.DistributedDataParallel(
            model, device_ids=[args.gpu], find_unused_parameters=True
        )
        model_without_ddp = model.module

    # get criterion
    if args.label_smooth > 0:
        criterion = LabelSmoothCrossEntropyLoss(num_classes=args.n_classes)
    else:
        criterion = nn.CrossEntropyLoss()
    logger.info("Criterion: {}".format(criterion))

    # get optimizer and scheduler
    optimizer = get_optimizer(model, args)
    scheduler = get_scheduler(optimizer, logger, args)
    logger.info("Optimizer: {}".format(optimizer))
    logger.info("Scheduler: {}".format(scheduler))

    if optimizer_state is not None:
        logger.info("Load optimizer state!")
        optimizer.load_state_dict(optimizer_state)
        logger.info("Current lr: {}".format(optimizer.param_groups[0]["lr"]))

    if lr_scheduler_state is not None:
        logger.info("Load lr state")
        scheduler.load_state_dict(lr_scheduler_state)
        logger.info(scheduler.last_epoch)
        logger.info(scheduler.get_last_lr())

    best_top1 = 100
    best_top5 = 100

    for epoch in range(start_epoch, args.n_epochs):
        if args.distributed and train_sampler is not None:
            train_sampler.set_epoch(epoch)

        # train for one epoch
        train_error, train_loss, train5_error = train(
            model,
            train_loader,
            criterion,
            optimizer,
            scheduler,
            device,
            logger,
            tensorboard_logger,
            epoch,
            args,
        )

        # evaluate on validation set
        val_error, val_loss, val5_error = val(
            model,
            val_loader,
            criterion,
            device,
            logger,
            tensorboard_logger,
            epoch,
            args,
        )

        # write log
        log_str = "{:d}\t".format(epoch)
        log_str += "{:.4f}\t{:.4f}\t{:.4f}\t{:.4f}\t{:.4f}\t{:.4f}\t".format(
            train_error, train_loss, val_error, val_loss, train5_error, val5_error
        )

        if args.rank == 0:
            write_log(args.save_path, "log.txt", log_str)

        # remember best acc@1, acc@5
        is_best = val_error <= best_top1
        best_top1 = min(best_top1, val_error)
        best_top5 = min(best_top5, val5_error)

        logger.info(
            "|===>Best Result is: Top1 Error: {:f}, Top5 Error: {:f}\n".format(
                best_top1, best_top5
            )
        )
        logger.info(
            "|==>Best Result is: Top1 Accuracy: {:f}, Top5 Accuracy: {:f}\n".format(
                100 - best_top1, 100 - best_top5
            )
        )

        # save checkpoint
        if args.rank == 0:
            if "imagenet" in args.dataset:
                checkpoint.save_checkpoint(
                    unwrap_model(model), optimizer, scheduler, epoch, epoch
                )
            else:
                checkpoint.save_checkpoint(
                    unwrap_model(model), optimizer, scheduler, epoch
                )

            if is_best:
                checkpoint.save_model(unwrap_model(model), best_flag=is_best)
